/*
 * Copyright (c) 1995, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2014-2020 MicroEJ Corp. - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package sun.net.www.protocol.http;

import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.HttpRetryException;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ej.annotation.Nullable;
import sun.net.NetworkClient;
import sun.net.www.MessageHeader;
import sun.net.www.http.ChunkedOutputStream;
import sun.net.www.http.HttpClient;
import sun.net.www.http.PosterOutputStream;

/**
 * A class to represent an HTTP connection to a remote object.
 */

public class HttpURLConnection extends java.net.HttpURLConnection {

	static String HTTP_CONNECT = "CONNECT";

	static final String version;
	public static final String userAgent;

	/* max # of allowed re-directs */
	static final int defaultmaxRedirects = 20;
	static final int maxRedirects;

	@Nullable
	private StreamingOutputStream strOutputStream;

	private final static String RETRY_MSG2 = "cannot retry due to server authentication, in streaming mode";
	private final static String RETRY_MSG3 = "cannot retry due to redirection, in streaming mode";
	/*
	 * System properties related to error stream handling:
	 *
	 * sun.net.http.errorstream.enableBuffering = <boolean>
	 *
	 * With the above system property set to true (default is false), when the response code is >=400, the HTTP handler
	 * will try to buffer the response body (up to a certain amount and within a time limit). Thus freeing up the
	 * underlying socket connection for reuse. The rationale behind this is that usually when the server responds with a
	 * >=400 error (client error or server error, such as 404 file not found), the server will send a small response
	 * body to explain who to contact and what to do to recover. With this property set to true, even if the application
	 * doesn't call getErrorStream(), read the response body, and then call close(), the underlying socket connection
	 * can still be kept-alive and reused. The following two system properties provide further control to the error
	 * stream buffering behaviour.
	 *
	 * sun.net.http.errorstream.timeout = <int> the timeout (in millisec) waiting the error stream to be buffered;
	 * default is 300 ms
	 *
	 * sun.net.http.errorstream.bufferSize = <int> the size (in bytes) to use for the buffering the error stream;
	 * default is 4k
	 */

	static {
		maxRedirects = Integer.getInteger("http.maxRedirects", defaultmaxRedirects).intValue();
		version = System.getProperty("java.version");
		String agent = System.getProperty(("http.agent"));
		if (agent == null) {
			agent = "Java/" + version;
		} else {
			agent = agent + " Java/" + version;
		}
		userAgent = agent;
	}

	static final String httpVersion = "HTTP/1.1";
	static final String acceptString = "text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2";

	@Nullable
	protected HttpClient http;
	protected Handler handler;

	@Nullable
	private MessageHeader cachedHeaders;
	@Nullable
	private InputStream cachedInputStream;

	/* output stream to server */
	@Nullable
	protected PrintStream ps = null;

	/* buffered error stream */
	@Nullable
	private final InputStream errorStream = null;

	/*
	 * all the headers we send NOTE: do *NOT* dump out the content of 'requests' in the output or stacktrace since it
	 * may contain security-sensitive headers such as those defined in EXCLUDE_HEADERS.
	 */
	private MessageHeader requests;

	/* The following two fields are only used with Digest Authentication */
	@Nullable
	String domain; /* The list of authentication domains */

	/* all the response headers we get back */
	private MessageHeader responses;
	/* the stream _from_ the server */
	@Nullable
	private InputStream inputStream = null;
	/* post stream _to_ the server, if any */
	@Nullable
	private PosterOutputStream poster = null;

	/* Indicates if the std. request headers have been set in requests. */
	private boolean setRequests = false;

	/* Indicates whether a request has already failed or not */
	private boolean failedOnce = false;

	/*
	 * Remembered Exception, we will throw it again if somebody calls getInputStream after disconnect
	 */
	@Nullable
	private Exception rememberedException = null;

	/* If we decide we want to reuse a client, we put it here */
	@Nullable
	private HttpClient reuseClient = null;

	/*
	 * Redefine timeouts from java.net.URLConnection as we need -1 to mean not set. This is to ensure backward
	 * compatibility.
	 */
	private int connectTimeout = NetworkClient.DEFAULT_CONNECT_TIMEOUT;
	private int readTimeout = NetworkClient.DEFAULT_READ_TIMEOUT;

	/*
	 * Checks the validity of http message header and whether the header is restricted and throws
	 * IllegalArgumentException if invalid or restricted.
	 */
	private boolean isExternalMessageHeaderAllowed(String key, String value) {
		checkMessageHeader(key, value);
		return true;
	}

	/*
	 * checks the validity of http message header and throws IllegalArgumentException if invalid.
	 */
	private void checkMessageHeader(String key, @Nullable String value) {
		char LF = '\n';
		int index = key.indexOf(LF);
		if (index != -1) {
			throw new IllegalArgumentException("Illegal character(s) in message header field: " + key);
		} else {
			if (value == null) {
				return;
			}

			index = value.indexOf(LF);
			while (index != -1) {
				index++;
				if (index < value.length()) {
					char c = value.charAt(index);
					if ((c == ' ') || (c == '\t')) {
						// ok, check the next occurrence
						index = value.indexOf(LF, index);
						continue;
					}
				}
				throw new IllegalArgumentException("Illegal character(s) in message header value: " + value);
			}
		}
	}

	/*
	 * adds the standard key/val pairs to reqests if necessary & write to given PrintStream
	 */
	private void writeRequests() throws IOException {
		/*
		 * print all message headers in the MessageHeader onto the wire - all the ones we've set and any others that
		 * have been set
		 */
		if (!this.setRequests) {

			/*
			 * We're very particular about the order in which we set the request headers here. The order should not
			 * matter, but some careless CGI programs have been written to expect a very particular order of the
			 * standard headers. To name names, the order in which Navigator3.0 sends them. In particular, we make
			 * *sure* to send Content-type: <> and Content-length:<> second to last and last, respectively, in the case
			 * of a POST request.
			 */
			if (!this.failedOnce) {
				this.requests.prepend(this.method + " " + getRequestURI() + " " + httpVersion, null);
			}
			if (!getUseCaches()) {
				this.requests.setIfNotSet("Cache-Control", "no-cache");
				this.requests.setIfNotSet("Pragma", "no-cache");
			}
			this.requests.setIfNotSet("User-Agent", userAgent);
			int port = this.url.getPort();
			String host = this.url.getHost();
			if (port != -1 && port != this.url.getDefaultPort()) {
				host += ":" + String.valueOf(port);
			}
			this.requests.setIfNotSet("Host", host);
			this.requests.setIfNotSet("Accept", acceptString);

			/*
			 * For HTTP/1.1 the default behavior is to keep connections alive. However, we may be talking to a 1.0
			 * server so we should set keep-alive just in case, except if we have encountered an error or if keep alive
			 * is disabled via a system property
			 */

			/*
			 * RFC 2616 HTTP/1.1 section 14.10 says: HTTP/1.1 applications that do not support persistent connections
			 * MUST include the "close" connection option in every message
			 */
			this.requests.setIfNotSet("Connection", "close");

			if (!this.method.equals("PUT") && (this.poster != null || streaming())) {
				this.requests.setIfNotSet("Content-type", "application/x-www-form-urlencoded");
			}

			boolean chunked = false;

			if (streaming()) {
				if (this.chunkLength != -1) {
					this.requests.set("Transfer-Encoding", "chunked");
					chunked = true;
				} else { /* fixed content length */
					if (this.fixedContentLengthLong != -1) {
						this.requests.set("Content-Length", String.valueOf(this.fixedContentLengthLong));
					} else if (this.fixedContentLength != -1) {
						this.requests.set("Content-Length", String.valueOf(this.fixedContentLength));
					}
				}
			} else if (this.poster != null) {
				/* add Content-Length & POST/PUT data */
				synchronized (this.poster) {
					/* close it, so no more data can be added */
					this.poster.close();
					this.requests.set("Content-Length", String.valueOf(this.poster.size()));
				}
			}

			if (!chunked) {
				if (this.requests.findValue("Transfer-Encoding") != null) {
					this.requests.remove("Transfer-Encoding");
				}
			}

			this.setRequests = true;
		}
		this.http.writeRequests(this.requests, this.poster, streaming());
		if (this.ps.checkError()) {
			disconnectInternal();
			if (this.failedOnce) {
				throw new IOException("Error writing to server");
			} else { // try once more
				this.failedOnce = true;
				setNewClient(this.url);
				this.ps = (PrintStream) this.http.getOutputStream();
				this.connected = true;
				this.responses = new MessageHeader();
				this.setRequests = false;
				writeRequests();
			}
		}
	}

	/**
	 * Obtain a HttpsClient object. Use the cached copy if specified.
	 *
	 * @param url
	 *            the URL being accessed
	 */
	protected void setNewClient(URL url) throws IOException {
		HttpClient http = this.http = HttpClient.New(url, this.connectTimeout, this);
		http.setReadTimeout(this.readTimeout);
	}

	/**
	 * this constructor is used by other protocol handlers such as ftp that want to use http to fetch urls on their
	 * behalf.
	 */
	public HttpURLConnection(URL u) {
		this(u, new Handler());
	}

	protected HttpURLConnection(URL u, Handler handler) {
		super(u);
		this.requests = new MessageHeader();
		this.responses = new MessageHeader();
		this.handler = handler;
	}

	/**
	 * opens a stream allowing redirects only to the same host.
	 */
	public static InputStream openConnectionCheckRedirects(URLConnection c) throws IOException {
		boolean redir;
		int redirects = 0;
		InputStream in;

		do {
			if (c instanceof HttpURLConnection) {
				((HttpURLConnection) c).setInstanceFollowRedirects(false);
			}

			// We want to open the input stream before
			// getting headers, because getHeaderField()
			// et al swallow IOExceptions.
			in = c.getInputStream();
			redir = false;

			if (c instanceof HttpURLConnection) {
				HttpURLConnection http = (HttpURLConnection) c;
				int stat = http.getResponseCode();
				if (stat >= 300 && stat <= 307 && stat != 306 && stat != HttpURLConnection.HTTP_NOT_MODIFIED) {
					URL base = http.getURL();
					String loc = http.getHeaderField("Location");
					URL target = null;
					if (loc != null) {
						target = new URL(base, loc);
					}
					http.disconnect();
					if (target == null || !base.getProtocol().equals(target.getProtocol())
							|| base.getPort() != target.getPort() || !hostsEqual(base, target) || redirects >= 5) {
						throw new SecurityException("illegal URL redirect");
					}
					redir = true;
					c = target.openConnection();
					redirects++;
				}
			}
		} while (redir);
		return in;
	}

	//
	// Same as java.net.URL.hostsEqual
	//
	private static boolean hostsEqual(URL u1, URL u2) {
		final String h1 = u1.getHost();
		final String h2 = u2.getHost();

		if (h1 == null) {
			return h2 == null;
		} else if (h2 == null) {
			return false;
		} else if (h1.equalsIgnoreCase(h2)) {
			return true;
		}
		// Have to resolve addresses before comparing, otherwise
		// names like tachyon and tachyon.eng would compare different
		final boolean result[] = { false };

		return result[0];
	}

	// overridden in HTTPS subclass

	@Override
	public void connect() throws IOException {
		plainConnect();
	}

	private boolean checkReuseConnection() {
		if (this.connected) {
			return true;
		}
		HttpClient reuseClient = this.reuseClient;
		if (reuseClient != null) {
			HttpClient http = this.http = reuseClient;
			http.setReadTimeout(getReadTimeout());
			http.reuse = false;
			this.reuseClient = null;
			this.connected = true;
			return true;
		}
		return false;
	}

	protected void plainConnect() throws IOException {
		if (this.connected) {
			return;
		}
		try {
			/*
			 * Try to open connections using the following scheme, return on the first one that's successful: 3) is 2)
			 * fails, make direct connection
			 */

			// No proxy selector, create http client with no proxy
			if (!this.failedOnce) {
				this.http = getNewHttpClient(this.url, this.connectTimeout);
				this.http.setReadTimeout(this.readTimeout);
			} else {
				// make sure to construct new connection if first
				// attempt failed
				this.http = getNewHttpClient(this.url, this.connectTimeout);
				this.http.setReadTimeout(this.readTimeout);
			}
			this.ps = (PrintStream) this.http.getOutputStream();
		} catch (IOException e) {
			throw e;
		}
		// constructor to HTTP client calls openserver
		this.connected = true;
	}

	// subclass HttpsClient will overwrite & return an instance of HttpsClient
	protected HttpClient getNewHttpClient(URL url, int connectTimeout) throws IOException {
		return HttpClient.New(url, connectTimeout, this);
	}

	private void expect100Continue() throws IOException {
		// Expect: 100-Continue was set, so check the return code for
		// Acceptance
		int oldTimeout = this.http.getReadTimeout();
		boolean enforceTimeOut = false;
		boolean timedOut = false;
		if (oldTimeout <= 0) {
			// 5s read timeout in case the server doesn't understand
			// Expect: 100-Continue
			this.http.setReadTimeout(5000);
			enforceTimeOut = true;
		}

		try {
			this.http.parseHTTP(this.responses, this);
		} catch (SocketTimeoutException se) {
			if (!enforceTimeOut) {
				throw se;
			}
			timedOut = true;
			this.http.setIgnoreContinue(true);
		}
		if (!timedOut) {
			// Can't use getResponseCode() yet
			String resp = this.responses.getValue(0);
			// Parse the response which is of the form:
			// HTTP/1.1 417 Expectation Failed
			// HTTP/1.1 100 Continue
			if (resp != null && resp.startsWith("HTTP/")) {
				String[] sa = HttpClient.splitWhitespace(resp);
				this.responseCode = -1;
				try {
					// Response code is 2nd token on the line
					if (sa.length > 1) {
						this.responseCode = Integer.parseInt(sa[1]);
					}
				} catch (NumberFormatException numberFormatException) {
				}
			}
			if (this.responseCode != 100) {
				throw new ProtocolException("Server rejected operation");
			}
		}

		this.http.setReadTimeout(oldTimeout);

		this.responseCode = -1;
		this.responses.reset();
		// Proceed
	}

	/*
	 * Allowable input/output sequences: [interpreted as POST/PUT] - get output, [write output,] get input, [read input]
	 * - get output, [write output] [interpreted as GET] - get input, [read input] Disallowed: - get input, [read
	 * input,] get output, [write output]
	 */

	@Override
	public synchronized OutputStream getOutputStream() throws IOException {

		try {
			if (!this.doOutput) {
				throw new ProtocolException(
						"cannot write to a URLConnection" + " if doOutput=false - call setDoOutput(true)");
			}

			if (this.method.equals("GET")) {
				this.method = "POST"; // Backward compatibility
			}
			if (!"POST".equals(this.method) && !"PUT".equals(this.method) && "http".equals(this.url.getProtocol())) {
				throw new ProtocolException("HTTP method " + this.method + " doesn't support output");
			}

			// if there's already an input stream open, throw an exception
			if (this.inputStream != null) {
				throw new ProtocolException("Cannot write output after reading input.");
			}

			if (!checkReuseConnection()) {
				connect();
			}

			boolean expectContinue = false;
			String expects = this.requests.findValue("Expect");
			if ("100-Continue".equalsIgnoreCase(expects)) {
				this.http.setIgnoreContinue(false);
				expectContinue = true;
			}

			if (streaming() && this.strOutputStream == null) {
				writeRequests();
			}

			if (expectContinue) {
				expect100Continue();
			}
			this.ps = (PrintStream) this.http.getOutputStream();
			if (streaming()) {
				if (this.strOutputStream == null) {
					if (this.chunkLength != -1) { /* chunked */
						this.strOutputStream = new StreamingOutputStream(
								new ChunkedOutputStream(this.ps, this.chunkLength), -1L);
					} else { /* must be fixed content length */
						long length = 0L;
						if (this.fixedContentLengthLong != -1) {
							length = this.fixedContentLengthLong;
						} else if (this.fixedContentLength != -1) {
							length = this.fixedContentLength;
						}
						this.strOutputStream = new StreamingOutputStream(this.ps, length);
					}
				}
				return this.strOutputStream;
			} else {
				PosterOutputStream poster = this.poster;
				if (poster == null) {
					poster = this.poster = new PosterOutputStream();
				}
				return poster;
			}
		} catch (RuntimeException e) {
			disconnectInternal();
			throw e;
		} catch (ProtocolException e) {
			// Save the response code which may have been set while enforcing
			// the 100-continue. disconnectInternal() forces it to -1
			int i = this.responseCode;
			disconnectInternal();
			this.responseCode = i;
			throw e;
		} catch (IOException e) {
			disconnectInternal();
			throw e;
		}
	}

	public boolean streaming() {
		return (this.fixedContentLength != -1) || (this.fixedContentLengthLong != -1) || (this.chunkLength != -1);
	}

	@Override
	public synchronized InputStream getInputStream() throws IOException {

		if (!this.doInput) {
			throw new ProtocolException("Cannot read from URLConnection" + " if doInput=false (call setDoInput(true))");
		}

		if (this.rememberedException != null) {
			if (this.rememberedException instanceof RuntimeException) {
				throw new RuntimeException(this.rememberedException);
			} else {
				throw new IOException(this.rememberedException);
			}
		}

		InputStream inputStream = this.inputStream;
		if (inputStream != null) {
			return inputStream;
		}

		if (streaming()) {
			if (this.strOutputStream == null) {
				getOutputStream();
			}
			/* make sure stream is closed */
			this.strOutputStream.close();
			if (!this.strOutputStream.writtenOK()) {
				throw new IOException("Incomplete output stream");
			}
		}

		int redirects = 0;
		int respCode = 0;
		long cl = -1;

		try {
			do {
				if (!checkReuseConnection()) {
					connect();
				}

				InputStream cachedInputStream = this.cachedInputStream;
				if (cachedInputStream != null) {
					return cachedInputStream;
				}

				/*
				 * REMIND: This exists to fix the HttpsURLConnection subclass. Hotjava needs to run on JDK1.1FCS. Do
				 * proper fix once a proper solution for SSL can be found.
				 */
				this.ps = (PrintStream) this.http.getOutputStream();

				if (!streaming()) {
					writeRequests();
				}
				this.http.parseHTTP(this.responses, this);

				this.inputStream = this.http.getInputStream();

				respCode = getResponseCode();
				if (respCode == -1) {
					disconnectInternal();
					throw new IOException("Invalid Http response");
				}
				if (respCode == HTTP_PROXY_AUTH) {

				} else {
					this.requests.remove("Proxy-Authorization");
				}

				if (respCode == HTTP_UNAUTHORIZED) {
					if (streaming()) {
						disconnectInternal();
						throw new HttpRetryException(RETRY_MSG2, HTTP_UNAUTHORIZED);
					}
				}

				this.requests.remove("Authorization");
				this.requests.remove("Proxy-Authorization");

				if (followRedirect()) {
					/*
					 * if we should follow a redirect, then the followRedirects() method will disconnect() and
					 * re-connect us to the new location
					 */
					redirects++;

					continue;
				}

				try {
					cl = Long.parseLong(this.responses.findValue("content-length"));
				} catch (Exception exc) {
				}
				;

				if (this.method.equals("HEAD") || cl == 0 || respCode == HTTP_NOT_MODIFIED
						|| respCode == HTTP_NO_CONTENT) {

					this.http.finished();
					this.http = null;
					this.inputStream = new EmptyInputStream();
					this.connected = false;
				}

				if (respCode == 200 || respCode == 203 || respCode == 206 || respCode == 300 || respCode == 301
						|| respCode == 410) {
				}

				if (respCode >= 400) {
					if (respCode == 404 || respCode == 410) {
						throw new IOException(this.url.toString());
					} else {
						throw new java.io.IOException("Server returned HTTP" + " response code: " + respCode
								+ " for URL: " + this.url.toString());
					}
				}
				this.poster = null;
				this.strOutputStream = null;
				return this.inputStream;
			} while (redirects < maxRedirects);

			throw new ProtocolException("Server redirected too many " + " times (" + redirects + ")");
		} catch (RuntimeException e) {
			disconnectInternal();
			this.rememberedException = e;
			throw e;
		} catch (IOException e) {
			this.rememberedException = e;
			throw e;
		} finally {
		}
	}

	@Override
	@Nullable
	public InputStream getErrorStream() {
		if (this.connected && this.responseCode >= 400) {
			// Client Error 4xx and Server Error 5xx
			if (this.errorStream != null) {
				return this.errorStream;
			} else if (this.inputStream != null) {
				return this.inputStream;
			}
		}
		return null;
	}

	static String connectRequestURI(URL url) {
		String host = url.getHost();
		int port = url.getPort();
		port = port != -1 ? port : url.getDefaultPort();

		return host + ":" + port;
	}

	/*
	 * The request URI used in the request line for this request. Also, needed for digest authentication
	 */
	@Nullable
	String requestURI = null;

	String getRequestURI() throws IOException {
		String requestURI = this.requestURI;
		if (requestURI == null) {
			requestURI = this.requestURI = this.http.getURLFile();
		}
		return requestURI;
	}

	/*
	 * Tells us whether to follow a redirect. If so, it closes the connection (break any keep-alive) and resets the url,
	 * re-connects, and resets the request property.
	 */
	private boolean followRedirect() throws IOException {
		if (!getInstanceFollowRedirects()) {
			return false;
		}

		int stat = getResponseCode();
		if (stat < 300 || stat > 307 || stat == 306 || stat == HTTP_NOT_MODIFIED) {
			return false;
		}
		String loc = getHeaderField("Location");
		if (loc == null) {
			/*
			 * this should be present - if not, we have no choice but to go forward w/ the response we got
			 */
			return false;
		}
		URL locUrl;
		try {
			locUrl = new URL(loc);
			if (!this.url.getProtocol().equalsIgnoreCase(locUrl.getProtocol())) {
				return false;
			}

		} catch (MalformedURLException mue) {
			// treat loc as a relative URI to conform to popular browsers
			locUrl = new URL(this.url, loc);
		}
		disconnectInternal();
		if (streaming()) {
			throw new HttpRetryException(RETRY_MSG3, stat, loc);
		}

		// clear out old response headers!!!!
		this.responses = new MessageHeader();

		// maintain previous headers, just change the name
		// of the file we're getting
		this.url = locUrl;
		this.requestURI = null; // force it to be recalculated
		if (this.method.equals("POST") && !Boolean.getBoolean("http.strictPostRedirect") && (stat != 307)) {
			/*
			 * The HTTP/1.1 spec says that a redirect from a POST *should not* be immediately turned into a GET, and
			 * that some HTTP/1.0 clients incorrectly did this. Correct behavior redirects a POST to another POST.
			 * Unfortunately, since most browsers have this incorrect behavior, the web works this way now. Typical
			 * usage seems to be: POST a login code or passwd to a web page. after validation, the server redirects to
			 * another (welcome) page The second request is (erroneously) expected to be GET
			 *
			 * We will do the incorrect thing (POST-->GET) by default. We will provide the capability to do the "right"
			 * thing (POST-->POST) by a system property, "http.strictPostRedirect=true"
			 */

			this.requests = new MessageHeader();
			this.setRequests = false;
			setRequestMethod("GET");
			this.poster = null;
			if (!checkReuseConnection()) {
				connect();
			}
		} else {
			if (!checkReuseConnection()) {
				connect();
			}
			/*
			 * Even after a connect() call, http variable still can be null, if a ResponseCache has been installed and
			 * it returns a non-null CacheResponse instance. So check nullity before using it.
			 *
			 * And further, if http is null, there's no need to do anything about request headers because successive
			 * http session will use cachedInputStream/cachedHeaders anyway, which is returned by CacheResponse.
			 */
			if (this.http != null) {
				this.requests.set(0, this.method + " " + getRequestURI() + " " + httpVersion, null);
				int port = this.url.getPort();
				String host = this.url.getHost();
				if (port != -1 && port != this.url.getDefaultPort()) {
					host += ":" + String.valueOf(port);
				}
				this.requests.set("Host", host);
			}
		}
		return true;
	}

	/**
	 * Disconnect from the server (for internal use)
	 */
	private void disconnectInternal() {
		this.responseCode = -1;
		this.inputStream = null;
		HttpClient http = this.http;
		if (http != null) {
			http.closeServer();
			this.http = null;
			this.connected = false;
		}
	}

	/**
	 * Disconnect from the server (public API)
	 */
	@Override
	public void disconnect() {

		this.responseCode = -1;
		HttpClient http = this.http;
		if (http != null) {
			/*
			 * If we have an input stream this means we received a response from the server. That stream may have been
			 * read to EOF and dependening on the stream type may already be closed or the the http client may be
			 * returned to the keep-alive cache. If the http client has been returned to the keep-alive cache it may be
			 * closed (idle timeout) or may be allocated to another request.
			 *
			 * In other to avoid timing issues we close the input stream which will either close the underlying
			 * connection or return the client to the cache. If there's a possibility that the client has been returned
			 * to the cache (ie: stream is a keep alive stream or a chunked input stream) then we remove an idle
			 * connection to the server. Note that this approach can be considered an approximation in that we may close
			 * a different idle connection to that used by the request. Additionally it's possible that we close two
			 * connections - the first becuase it wasn't an EOF (and couldn't be hurried) - the second, another idle
			 * connection to the same server. The is okay because "disconnect" is an indication that the application
			 * doesn't intend to access this http server for a while.
			 */

			InputStream inputStream = this.inputStream;
			if (inputStream != null) {
				try {
					inputStream.close();
				} catch (IOException ioe) {
				}
			} else {
				// We are deliberatly being disconnected so HttpClient
				// should not try to resend the request no matter what stage
				// of the connection we are in.
				http.setDoNotRetry(true);

				http.closeServer();
			}

			// poster = null;
			this.http = null;
			this.connected = false;
		}
		this.cachedInputStream = null;
		MessageHeader cachedHeaders = this.cachedHeaders;
		if (cachedHeaders != null) {
			cachedHeaders.reset();
		}
	}

	@Override
	public boolean usingProxy() {
		return false;
	}

	/**
	 * Returns a filtered version of the given headers value.
	 *
	 * Note: The implementation currently only filters out HttpOnly cookies from Set-Cookie and Set-Cookie2 headers.
	 */
	@Nullable
	private String filterHeaderField(@Nullable String name, @Nullable String value) {
		if (value == null) {
			return null;
		}

		return value;
	}

	// Cache the filtered response headers so that they don't need
	// to be generated for every getHeaderFields() call.
	@Nullable
	private Map<String, List<String>> filteredHeaders; // null

	private Map<String, List<String>> getFilteredHeaderFields() {
		Map<String, List<String>> filteredHeaders = this.filteredHeaders;
		if (filteredHeaders != null) {
			return filteredHeaders;
		}

		Map<String, List<String>> headers, tmpMap = new HashMap<>();

		MessageHeader cachedHeaders = this.cachedHeaders;
		if (cachedHeaders != null) {
			headers = cachedHeaders.getHeaders();
		} else {
			headers = this.responses.getHeaders();
		}

		for (Map.Entry<String, List<String>> e : headers.entrySet()) {
			String key = e.getKey();
			List<String> values = e.getValue(), filteredVals = new ArrayList<>();
			for (String value : values) {
				String fVal = filterHeaderField(key, value);
				if (fVal != null) {
					filteredVals.add(fVal);
				}
			}
			if (!filteredVals.isEmpty()) {
				tmpMap.put(key, Collections.unmodifiableList(filteredVals));
			}
		}

		return this.filteredHeaders = Collections.unmodifiableMap(tmpMap);
	}

	/**
	 * Gets a header field by name. Returns null if not known.
	 *
	 * @param name
	 *            the name of the header field
	 */
	@Override
	@Nullable
	public String getHeaderField(String name) {
		try {
			getInputStream();
		} catch (IOException e) {
		}

		MessageHeader cachedHeaders = this.cachedHeaders;
		if (cachedHeaders != null) {
			return filterHeaderField(name, cachedHeaders.findValue(name));
		}

		return filterHeaderField(name, this.responses.findValue(name));
	}

	/**
	 * Returns an unmodifiable Map of the header fields. The Map keys are Strings that represent the response-header
	 * field names. Each Map value is an unmodifiable List of Strings that represents the corresponding field values.
	 *
	 * @return a Map of header fields
	 * @since 1.4
	 */
	@Override
	public Map<String, List<String>> getHeaderFields() {
		try {
			getInputStream();
		} catch (IOException e) {
		}

		return getFilteredHeaderFields();
	}

	/**
	 * Gets a header field by index. Returns null if not known.
	 *
	 * @param n
	 *            the index of the header field
	 */
	@Override
	@Nullable
	public String getHeaderField(int n) {
		try {
			getInputStream();
		} catch (IOException e) {
		}

		MessageHeader cachedHeaders = this.cachedHeaders;
		if (cachedHeaders != null) {
			return filterHeaderField(cachedHeaders.getKey(n), cachedHeaders.getValue(n));
		}
		return filterHeaderField(this.responses.getKey(n), this.responses.getValue(n));
	}

	/**
	 * Gets a header field by index. Returns null if not known.
	 *
	 * @param n
	 *            the index of the header field
	 */
	@Override
	@Nullable
	public String getHeaderFieldKey(int n) {
		try {
			getInputStream();
		} catch (IOException e) {
		}

		MessageHeader cachedHeaders = this.cachedHeaders;
		if (cachedHeaders != null) {
			return cachedHeaders.getKey(n);
		}

		return this.responses.getKey(n);
	}

	/**
	 * Sets request property. If a property with the key already exists, overwrite its value with the new value.
	 *
	 * @param value
	 *            the value to be set
	 */
	@Override
	public void setRequestProperty(String key, String value) {
		if (this.connected) {
			throw new IllegalStateException("Already connected");
		}
		if (key == null) {
			throw new NullPointerException("key is null");
		}

		if (isExternalMessageHeaderAllowed(key, value)) {
			this.requests.set(key, value);
		}
	}

	/**
	 * Adds a general request property specified by a key-value pair. This method will not overwrite existing values
	 * associated with the same key.
	 *
	 * @param key
	 *            the keyword by which the request is known (e.g., "<code>accept</code>").
	 * @param value
	 *            the value associated with it.
	 * @see #getRequestProperties()
	 * @since 1.4
	 */
	@Override
	public void addRequestProperty(String key, String value) {
		if (this.connected) {
			throw new IllegalStateException("Already connected");
		}
		if (key == null) {
			throw new NullPointerException("key is null");
		}

		if (isExternalMessageHeaderAllowed(key, value)) {
			this.requests.add(key, value);
		}
	}

	//
	// Set a property for authentication. This can safely disregard
	// the connected test.
	//
	public void setAuthenticationProperty(String key, String value) {
		checkMessageHeader(key, value);
		this.requests.set(key, value);
	}

	@Override
	@Nullable
	public synchronized String getRequestProperty(@Nullable String key) {
		if (key == null) {
			return null;
		}
		return this.requests.findValue(key);
	}

	/**
	 * Returns an unmodifiable Map of general request properties for this connection. The Map keys are Strings that
	 * represent the request-header field names. Each Map value is a unmodifiable List of Strings that represents the
	 * corresponding field values.
	 *
	 * @return a Map of the general request properties for this connection.
	 * @throws IllegalStateException
	 *             if already connected
	 * @since 1.4
	 */
	@Override
	public synchronized Map<String, List<String>> getRequestProperties() {
		if (this.connected) {
			throw new IllegalStateException("Already connected");
		}

		return this.requests.filterAndAddHeaders(null, null);
	}

	@Override
	public void setConnectTimeout(int timeout) {
		if (timeout < 0) {
			throw new IllegalArgumentException("timeouts can't be negative");
		}
		this.connectTimeout = timeout;
	}

	/**
	 * Returns setting for connect timeout.
	 * <p>
	 * 0 return implies that the option is disabled (i.e., timeout of infinity).
	 *
	 * @return an <code>int</code> that indicates the connect timeout value in milliseconds
	 * @see java.net.URLConnection#setConnectTimeout(int)
	 * @see java.net.URLConnection#connect()
	 * @since 1.5
	 */
	@Override
	public int getConnectTimeout() {
		return (this.connectTimeout < 0 ? 0 : this.connectTimeout);
	}

	/**
	 * Sets the read timeout to a specified timeout, in milliseconds. A non-zero value specifies the timeout when
	 * reading from Input stream when a connection is established to a resource. If the timeout expires before there is
	 * data available for read, a java.net.SocketTimeoutException is raised. A timeout of zero is interpreted as an
	 * infinite timeout.
	 *
	 * <p>
	 * Some non-standard implementation of this method ignores the specified timeout. To see the read timeout set,
	 * please call getReadTimeout().
	 *
	 * @param timeout
	 *            an <code>int</code> that specifies the timeout value to be used in milliseconds
	 * @throws IllegalArgumentException
	 *             if the timeout parameter is negative
	 *
	 * @see java.io.InputStream#read()
	 * @since 1.5
	 */
	@Override
	public void setReadTimeout(int timeout) {
		if (timeout < 0) {
			throw new IllegalArgumentException("timeouts can't be negative");
		}
		this.readTimeout = timeout;
	}

	/**
	 * Returns setting for read timeout. 0 return implies that the option is disabled (i.e., timeout of infinity).
	 *
	 * @return an <code>int</code> that indicates the read timeout value in milliseconds
	 *
	 * @see java.net.URLConnection#setReadTimeout(int)
	 * @see java.io.InputStream#read()
	 * @since 1.5
	 */
	@Override
	public int getReadTimeout() {
		return this.readTimeout < 0 ? 0 : this.readTimeout;
	}

	String getMethod() {
		return this.method;
	}

	class StreamingOutputStream extends FilterOutputStream {

		long expected;
		long written;
		boolean closed;
		boolean error;
		@Nullable
		IOException errorExcp;

		/**
		 * expectedLength == -1 if the stream is chunked expectedLength > 0 if the stream is fixed content-length In the
		 * 2nd case, we make sure the expected number of of bytes are actually written
		 */
		StreamingOutputStream(OutputStream os, long expectedLength) {
			super(os);
			this.expected = expectedLength;
			this.written = 0L;
			this.closed = false;
			this.error = false;
		}

		@Override
		public void write(int b) throws IOException {
			checkError();
			this.written++;
			if (this.expected != -1L && this.written > this.expected) {
				throw new IOException("too many bytes written");
			}
			this.out.write(b);
		}

		@Override
		public void write(byte[] b) throws IOException {
			write(b, 0, b.length);
		}

		@Override
		public void write(byte[] b, int off, int len) throws IOException {
			checkError();
			this.written += len;
			if (this.expected != -1L && this.written > this.expected) {
				this.out.close();
				throw new IOException("too many bytes written");
			}
			this.out.write(b, off, len);
		}

		void checkError() throws IOException {
			if (this.closed) {
				throw new IOException("Stream is closed");
			}
			if (this.error) {
				throw this.errorExcp;
			}
			if (((PrintStream) this.out).checkError()) {
				throw new IOException("Error writing request body to server");
			}
		}

		/*
		 * this is called to check that all the bytes that were supposed to be written were written and that the stream
		 * is now closed().
		 */
		boolean writtenOK() {
			return this.closed && !this.error;
		}

		@Override
		public void close() throws IOException {
			if (this.closed) {
				return;
			}
			this.closed = true;
			if (this.expected != -1L) {
				/* not chunked */
				if (this.written != this.expected) {
					this.error = true;
					IOException errorExcp = this.errorExcp = new IOException("insufficient data written");
					this.out.close();
					throw errorExcp;
				}
				super.flush(); /* can't close the socket */
			} else {
				/* chunked */
				super.close(); /* force final chunk to be written */
				/* trailing \r\n */
				OutputStream o = HttpURLConnection.this.http.getOutputStream();
				o.write('\r');
				o.write('\n');
				o.flush();
			}
		}
	}
}

/**
 * An input stream that just returns EOF. This is for HTTP URLConnections that are KeepAlive && use the HEAD method -
 * i.e., stream not dead, but nothing to be read.
 */

class EmptyInputStream extends InputStream {

	@Override
	public int available() {
		return 0;
	}

	@Override
	public int read() {
		return -1;
	}
}
